@@ -26,6 +26,7 @@
 
 #include <math.h>
 #include <string.h>
+#include <stdarg.h>
 
 #include <libart_lgpl/art_affine.h>
 #include <libart_lgpl/art_vpath_bpath.h>
@@ -51,6 +52,7 @@
 typedef struct _RsvgState RsvgState;
 typedef struct _RsvgSaxHandler RsvgSaxHandler;
 
+#define SVG_BUFFER_SIZE (1024 * 8)
 
 struct _RsvgState {
   double affine[6];
@@ -158,6 +160,12 @@ rsvg_ctx_free_helper (gpointer key, gpointer value, gpointer user_data)
   g_free (entval);
 }
 
+static void
+rsvg_pixmap_destroy (guchar *pixels, gpointer data)
+{
+  g_free (pixels);
+}
+
 static void
 rsvg_start_svg (RsvgHandle *ctx, const xmlChar **atts)
 {
@@ -191,29 +199,47 @@ rsvg_start_svg (RsvgHandle *ctx, const xmlChar **atts)
       if (ctx->size_func)
 	(* ctx->size_func) (&new_width, &new_height, ctx->user_data);
 
-      x_zoom = ((gdouble)new_width)/width;
-      y_zoom = ((gdouble)new_height)/height;
+      x_zoom = (width < 0 || new_width < 0) ? 1 : (double) new_width / width;
+      y_zoom = (height < 0 || new_height < 0) ? 1 : (double) new_height / height;
 
-      /* FIXME: Add GError here if size is wrong */
       /* Scale size of target pixbuf */
       state = &ctx->state[ctx->n_state - 1];
       art_affine_scale (state->affine, x_zoom, y_zoom);
 
+      if (new_width < 0 || new_height < 0)
+        {
+          g_warning ("rsvg_start_svg: no width and height attributes in SVG, nor supplied by size_func");
+          if (new_width < 0) new_width = 500;
+          if (new_height < 0) new_height = 500;
+        }
+
       if (new_width >= INT_MAX / 4)
-	return;
+        {
+          /* FIXME: What warning, GError here? */
+	  return;
+        }
       rowstride = (new_width * (has_alpha ? 4 : 3) + 3) & ~3;
       if (rowstride > INT_MAX / new_height)
-	return;
+        {
+          /* FIXME: What warning, GError here? */
+	  return;
+        }
+
+      /* FIXME: Add GError here if size is too big. */
+
       pixels = g_try_malloc (rowstride * new_height);
       if (pixels == NULL)
-	return;
+        {
+          /* FIXME: What warning, GError here? */
+	  return;
+        }
       memset (pixels, has_alpha ? 0 : 255, rowstride * new_height);
       ctx->pixbuf = gdk_pixbuf_new_from_data (pixels,
 					      GDK_COLORSPACE_RGB,
 					      has_alpha, 8,
 					      new_width, new_height,
 					      rowstride,
-					      NULL,
+					      rsvg_pixmap_destroy,
 					      NULL);
     }
 }
@@ -519,14 +545,20 @@ rsvg_push_opacity_group (RsvgHandle *ctx)
   state = &ctx->state[ctx->n_state - 1];
   pixbuf = ctx->pixbuf;
 
+  state->save_pixbuf = pixbuf;
+
+  if (pixbuf == NULL)
+    {
+      /* FIXME: What warning/GError here? */
+      return;
+    }
+
   if (!gdk_pixbuf_get_has_alpha (pixbuf))
     {
       g_warning ("push/pop transparency group on non-alpha buffer nyi");
       return;
     }
 
-  state->save_pixbuf = pixbuf;
-
   width = gdk_pixbuf_get_width (pixbuf);
   height = gdk_pixbuf_get_height (pixbuf);
   rowstride = gdk_pixbuf_get_rowstride (pixbuf);
@@ -540,7 +572,7 @@ rsvg_push_opacity_group (RsvgHandle *ctx)
 				     width,
 				     height,
 				     rowstride,
-				     NULL,
+				     rsvg_pixmap_destroy,
 				     NULL);
   ctx->pixbuf = pixbuf;
 }
@@ -568,6 +600,12 @@ rsvg_pop_opacity_group (RsvgHandle *ctx, int opacity)
   tos = ctx->pixbuf;
   nos = state->save_pixbuf;
 
+  if (tos == NULL || nos == NULL)
+    {
+      /* FIXME: What warning/GError here? */
+      return;
+    }
+
   if (!gdk_pixbuf_get_has_alpha (nos))
     {
       g_warning ("push/pop transparency group on non-alpha buffer nyi");
@@ -705,12 +743,11 @@ rsvg_render_svp (RsvgHandle *ctx, const ArtSVP *svp,
   gboolean has_alpha;
 
   pixbuf = ctx->pixbuf;
-  /* if a pixbuf hasn't been allocated, the svg is probably misformed.  Exit
-   * to avoid crashing.
-   */
-  if (pixbuf == NULL) {
-  	return;
-  }
+  if (pixbuf == NULL)
+    {
+      /* FIXME: What warning/GError here? */
+      return;
+    }
 
   has_alpha = gdk_pixbuf_get_has_alpha (pixbuf);
 
@@ -743,8 +780,14 @@ rsvg_render_bpath (RsvgHandle *ctx, const ArtBpath *bpath)
   int opacity;
   int tmp;
 
-  state = &ctx->state[ctx->n_state - 1];
   pixbuf = ctx->pixbuf;
+  if (pixbuf == NULL)
+    {
+      /* FIXME: What warning/GError here? */
+      return;
+    }
+
+  state = &ctx->state[ctx->n_state - 1];
   affine_bpath = art_bpath_affine_transform (bpath,
 					     state->affine);
 
@@ -873,6 +916,19 @@ rsvg_text_handler_characters (RsvgSaxHandler *self, const xmlChar *ch, int len)
   PangoRectangle ink_rect, line_ink_rect;
   FT_Bitmap bitmap;
 
+  state = &ctx->state[ctx->n_state - 1];
+  if (state->fill == NULL && state->font_size <= 0)
+    {
+      return;
+    }
+
+  pixbuf = ctx->pixbuf;
+  if (pixbuf == NULL)
+    {
+      /* FIXME: What warning/GError here? */
+      return;
+    }
+
   /* Copy ch into string, chopping off leading and trailing whitespace */
   for (beg = 0; beg < len; beg++)
     if (!g_ascii_isspace (ch[beg]))
@@ -893,68 +949,62 @@ rsvg_text_handler_characters (RsvgSaxHandler *self, const xmlChar *ch, int len)
   if (ctx->pango_context == NULL)
     ctx->pango_context = pango_ft2_get_context (72, 72); /* FIXME: dpi? */
 
-  state = &ctx->state[ctx->n_state - 1];
-
-  if (state->fill != NULL && state->font_size > 0)
-    {
-      pixbuf = ctx->pixbuf;
-      has_alpha = gdk_pixbuf_get_has_alpha (pixbuf);
-
-      render = art_render_new (0, 0,
-			       gdk_pixbuf_get_width (pixbuf),
-			       gdk_pixbuf_get_height (pixbuf),
-			       gdk_pixbuf_get_pixels (pixbuf),
-			       gdk_pixbuf_get_rowstride (pixbuf),
-			       gdk_pixbuf_get_n_channels (pixbuf) -
-			       (has_alpha ? 1 : 0),
-			       gdk_pixbuf_get_bits_per_sample (pixbuf),
-			       has_alpha ? ART_ALPHA_SEPARATE : ART_ALPHA_NONE,
-			       NULL);
-
-      layout = pango_layout_new (ctx->pango_context);
-      pango_layout_set_text (layout, string, -1);
-      font = pango_font_description_copy (pango_context_get_font_description (ctx->pango_context));
-      if (state->font_family)
-	pango_font_description_set_family_static (font, state->font_family);
-      pango_font_description_set_size (font, state->font_size * PANGO_SCALE);
-      pango_layout_set_font_description (layout, font);
-      pango_font_description_free (font);
-
-      pango_layout_get_pixel_extents (layout, &ink_rect, NULL);
-
-      line = pango_layout_get_line (layout, 0);
-      if (line == NULL)
-	line_ink_rect = ink_rect; /* nothing to draw anyway */
-      else
-	pango_layout_line_get_pixel_extents (line, &line_ink_rect, NULL);
-      
-      bitmap.rows = ink_rect.height;
-      bitmap.width = ink_rect.width;
-      bitmap.pitch = (bitmap.width + 3) & ~3;
-      bitmap.buffer = g_malloc0 (bitmap.rows * bitmap.pitch);
-      bitmap.num_grays = 0x100;
-      bitmap.pixel_mode = ft_pixel_mode_grays;
-
-      pango_ft2_render_layout (&bitmap, layout, -ink_rect.x, -ink_rect.y);
-
-      g_object_unref (layout);
-
-      rsvg_render_paint_server (render, state->fill, NULL); /* todo: paint server ctx */
-      opacity = state->fill_opacity * state->opacity;
-      opacity = opacity + (opacity >> 7) + (opacity >> 14);
+  has_alpha = gdk_pixbuf_get_has_alpha (pixbuf);
+  
+  render = art_render_new (0, 0,
+			   gdk_pixbuf_get_width (pixbuf),
+			   gdk_pixbuf_get_height (pixbuf),
+			   gdk_pixbuf_get_pixels (pixbuf),
+			   gdk_pixbuf_get_rowstride (pixbuf),
+			   gdk_pixbuf_get_n_channels (pixbuf) -
+			   (has_alpha ? 1 : 0),
+			   gdk_pixbuf_get_bits_per_sample (pixbuf),
+			   has_alpha ? ART_ALPHA_SEPARATE : ART_ALPHA_NONE,
+			   NULL);
+  
+  layout = pango_layout_new (ctx->pango_context);
+  pango_layout_set_text (layout, string, -1);
+  font = pango_font_description_copy (pango_context_get_font_description (ctx->pango_context));
+  if (state->font_family)
+    pango_font_description_set_family_static (font, state->font_family);
+  pango_font_description_set_size (font, state->font_size * PANGO_SCALE);
+  pango_layout_set_font_description (layout, font);
+  pango_font_description_free (font);
+  
+  pango_layout_get_pixel_extents (layout, &ink_rect, NULL);
+  
+  line = pango_layout_get_line (layout, 0);
+  if (line == NULL)
+    line_ink_rect = ink_rect; /* nothing to draw anyway */
+  else
+    pango_layout_line_get_pixel_extents (line, &line_ink_rect, NULL);
+  
+  bitmap.rows = ink_rect.height;
+  bitmap.width = ink_rect.width;
+  bitmap.pitch = (bitmap.width + 3) & ~3;
+  bitmap.buffer = g_malloc0 (bitmap.rows * bitmap.pitch);
+  bitmap.num_grays = 0x100;
+  bitmap.pixel_mode = ft_pixel_mode_grays;
+  
+  pango_ft2_render_layout (&bitmap, layout, -ink_rect.x, -ink_rect.y);
+  
+  g_object_unref (layout);
+  
+  rsvg_render_paint_server (render, state->fill, NULL); /* todo: paint server ctx */
+  opacity = state->fill_opacity * state->opacity;
+  opacity = opacity + (opacity >> 7) + (opacity >> 14);
 #ifdef VERBOSE
-      fprintf (stderr, "opacity = %d\n", opacity);
+  fprintf (stderr, "opacity = %d\n", opacity);
 #endif
-      art_render_mask_solid (render, opacity);
-      art_render_mask (render,
-		       state->affine[4] + line_ink_rect.x,
-		       state->affine[5] + line_ink_rect.y,
-		       state->affine[4] + line_ink_rect.x + bitmap.width,
-		       state->affine[5] + line_ink_rect.y + bitmap.rows,
-		       bitmap.buffer, bitmap.pitch);
-      art_render_invoke (render);
-      g_free (bitmap.buffer);
-    }
+  art_render_mask_solid (render, opacity);
+  art_render_mask (render,
+		   state->affine[4] + line_ink_rect.x,
+		   state->affine[5] + line_ink_rect.y,
+		   state->affine[4] + line_ink_rect.x + bitmap.width,
+		   state->affine[5] + line_ink_rect.y + bitmap.rows,
+		   bitmap.buffer, bitmap.pitch);
+  art_render_invoke (render);
+  g_free (bitmap.buffer);
 
   g_free (string);
 }
@@ -1327,6 +1377,19 @@ rsvg_entity_decl (void *data, const xmlChar *name, int type,
   g_hash_table_insert (entities, dupname, entity);
 }
 
+static void
+rsvg_error_cb (void *data, const char *msg, ...)
+{
+	va_list args;
+	
+	va_start (args, msg);
+#ifdef VERBOSE
+	fprintf (stderr, "fatal svg parse error: ");
+#endif
+	vfprintf (stderr, msg, args);
+	va_end (args);
+}
+
 static xmlSAXHandler rsvgSAXHandlerStruct = {
     NULL, /* internalSubset */
     NULL, /* isStandalone */
@@ -1350,13 +1413,11 @@ static xmlSAXHandler rsvgSAXHandlerStruct = {
     NULL, /* processingInstruction */
     NULL, /* comment */
     NULL, /* xmlParserWarning */
-    NULL, /* xmlParserError */
-    NULL, /* xmlParserError */
+    rsvg_error_cb, /* xmlParserError */
+    rsvg_error_cb, /* xmlParserFatalError */
     NULL, /* getParameterEntity */
 };
 
-static xmlSAXHandlerPtr rsvgSAXHandler = &rsvgSAXHandlerStruct;
-
 GQuark
 rsvg_error_quark (void)
 {
@@ -1452,16 +1513,13 @@ rsvg_handle_write (RsvgHandle    *handle,
   handle->error = &real_error;
   if (handle->ctxt == NULL)
     {
-      handle->ctxt = xmlCreatePushParserCtxt (rsvgSAXHandler,
-					      handle,
-					      buf, count,
-					      "filename XXX");
+      handle->ctxt = xmlCreatePushParserCtxt (
+	      &rsvgSAXHandlerStruct, handle, NULL, 0, NULL);
       handle->ctxt->replaceEntities = TRUE;
     }
-  else
-    {
-      xmlParseChunk (handle->ctxt, buf, count, 0);
-    }
+
+  xmlParseChunk (handle->ctxt, buf, count, 0);
+
   handle->error = NULL;
   /* FIXME: Error handling not implemented. */
   /*  if (*real_error != NULL)
@@ -1487,11 +1545,11 @@ gboolean
 rsvg_handle_close (RsvgHandle  *handle,
 		   GError     **error)
 {
-  gchar chars[1];
+  gchar chars[1] = { '\0' };
   GError *real_error;
 
   handle->error = &real_error;
-  xmlParseChunk (handle->ctxt, chars, 0, 1);
+  xmlParseChunk (handle->ctxt, chars, 1, TRUE);
   xmlFreeParserCtxt (handle->ctxt);
   /* FIXME: Error handling not implemented. */
   /*
@@ -1555,13 +1613,22 @@ rsvg_handle_free (RsvgHandle *handle)
   g_free (handle);
 }
 
+typedef enum {
+  RSVG_SIZE_ZOOM,
+  RSVG_SIZE_WH,
+  RSVG_SIZE_WH_MAX,
+  RSVG_SIZE_ZOOM_MAX
+} RsvgSizeType;
+
 struct RsvgSizeCallbackData
 {
   gdouble x_zoom;
   gdouble y_zoom;
   gint width;
   gint height;
+  RsvgSizeType type;
   gboolean zoom_set;
+  gboolean max_size_set;
 };
 
 static void
@@ -1570,19 +1637,42 @@ rsvg_size_callback (gint     *width,
 		    gpointer  data)
 {
   struct RsvgSizeCallbackData *real_data = (struct RsvgSizeCallbackData *)data;
+  gdouble zoomx, zoomy, zoom;
 
-  if (real_data->zoom_set)
-    {
-      (* width) = real_data->x_zoom * (* width);
-      (* height) = real_data->y_zoom * (* height);
-    }
-  else
-    {
-      if (real_data->width != -1)
-	*width = real_data->width;
-      if (real_data->height != -1)
-	*height = real_data->height;
-    }
+  switch (real_data->type) {
+  case RSVG_SIZE_ZOOM:
+	  (* width) = floor (real_data->x_zoom * (* width) + 0.5);
+	  (* height) = floor (real_data->y_zoom * (* height) + 0.5);
+	  break;
+  case RSVG_SIZE_ZOOM_MAX:
+	  (* width) = floor (real_data->x_zoom * (* width) + 0.5);
+	  (* height) = floor (real_data->y_zoom * (* height) + 0.5);
+
+	  if (*width > real_data->width || *height > real_data->height) {
+		  zoomx = (gdouble) real_data->width / (* width);
+		  zoomy = (gdouble) real_data->height / (* height);
+		  zoom = MIN (zoomx, zoomy);
+		  
+		  (* width) = floor (zoom * (* width) + 0.5);
+		  (* height) = floor (zoom * (* height) + 0.5);
+	  }
+	  break;
+  case RSVG_SIZE_WH_MAX:
+	  zoomx = (gdouble) real_data->width / (* width);
+	  zoomy = (gdouble) real_data->height / (* height);
+	  zoom = MIN (zoomx, zoomy);
+	  
+	  (* width) = floor (zoom * (* width) + 0.5);
+	  (* height) = floor (zoom * (* height) + 0.5);
+	  break;
+  case RSVG_SIZE_WH:
+  default:
+	  if (real_data->width != -1)
+		  *width = real_data->width;
+	  if (real_data->height != -1)
+		  *height = real_data->height;
+	  break;
+  }
 }
 
 /**
@@ -1603,7 +1693,6 @@ rsvg_pixbuf_from_file (const gchar *file_name,
   return rsvg_pixbuf_from_file_at_size (file_name, -1, -1, error);
 }
 
-
 /**
  * rsvg_pixbuf_from_file_at_zoom:
  * @file_name: A file name
@@ -1613,7 +1702,7 @@ rsvg_pixbuf_from_file (const gchar *file_name,
  * 
  * Loads a new #GdkPixbuf from @file_name and returns it.  This pixbuf is scaled
  * from the size indicated by the file by a factor of @x_zoom and @y_zoom.  The
- * caller must assume the reference to the reurned pixbuf. If an error
+ * caller must assume the reference to the returned pixbuf. If an error
  * occurred, @error is set and %NULL is returned.
  * 
  * Return value: A newly allocated #GdkPixbuf, or %NULL
@@ -1625,7 +1714,7 @@ rsvg_pixbuf_from_file_at_zoom (const gchar *file_name,
 			       GError     **error)
 {
   FILE *f;
-  char chars[10];
+  char chars[SVG_BUFFER_SIZE];
   gint result;
   GdkPixbuf *retval;
   RsvgHandle *handle;
@@ -1642,14 +1731,76 @@ rsvg_pixbuf_from_file_at_zoom (const gchar *file_name,
     }
 
   handle = rsvg_handle_new ();
-  data.zoom_set = TRUE;
+  data.type = RSVG_SIZE_ZOOM;
   data.x_zoom = x_zoom;
   data.y_zoom = y_zoom;
 
   rsvg_handle_set_size_callback (handle, rsvg_size_callback, &data, NULL);
-  while ((result = fread (chars, 1, 3, f)) > 0)
+  while ((result = fread (chars, 1, SVG_BUFFER_SIZE, f)) > 0)
     rsvg_handle_write (handle, chars, result, error);
+  rsvg_handle_close (handle, error);
+  
+  retval = rsvg_handle_get_pixbuf (handle);
 
+  fclose (f);
+  rsvg_handle_free (handle);
+  return retval;
+}
+
+/**
+ * rsvg_pixbuf_from_file_at_zoom_with_max:
+ * @file_name: A file name
+ * @x_zoom: The horizontal zoom factor
+ * @y_zoom: The vertical zoom factor
+ * @max_width: The requested max width
+ * @max_height: The requested max heigh
+ * @error: return location for errors
+ * 
+ * Loads a new #GdkPixbuf from @file_name and returns it.  This pixbuf is scaled
+ * from the size indicated by the file by a factor of @x_zoom and @y_zoom. If the
+ * resulting pixbuf would be larger than max_width/max_heigh it is uniformly scaled
+ * down to fit in that rectangle.The caller must assume the reference to the
+ * returned pixbuf. If an error occurred, @error is set and %NULL is returned.
+ * 
+ * Return value: A newly allocated #GdkPixbuf, or %NULL
+ **/
+GdkPixbuf  *
+rsvg_pixbuf_from_file_at_zoom_with_max (const gchar  *file_name,
+					double        x_zoom,
+					double        y_zoom,
+					gint          max_width,
+					gint          max_height,
+					GError      **error)
+{
+  FILE *f;
+  char chars[SVG_BUFFER_SIZE];
+  gint result;
+  GdkPixbuf *retval;
+  RsvgHandle *handle;
+  struct RsvgSizeCallbackData data;
+
+  g_return_val_if_fail (file_name != NULL, NULL);
+  g_return_val_if_fail (x_zoom > 0.0 && y_zoom > 0.0, NULL);
+
+  f = fopen (file_name, "r");
+  if (!f)
+    {
+      /* FIXME: Set up error. */
+      return NULL;
+    }
+
+  handle = rsvg_handle_new ();
+  data.type = RSVG_SIZE_ZOOM_MAX;
+  data.x_zoom = x_zoom;
+  data.y_zoom = y_zoom;
+  data.width = max_width;
+  data.height = max_height;
+
+  rsvg_handle_set_size_callback (handle, rsvg_size_callback, &data, NULL);
+  while ((result = fread (chars, 1, SVG_BUFFER_SIZE, f)) > 0)
+    rsvg_handle_write (handle, chars, result, error);
+  rsvg_handle_close (handle, error);
+  
   retval = rsvg_handle_get_pixbuf (handle);
 
   fclose (f);
@@ -1667,7 +1818,7 @@ rsvg_pixbuf_from_file_at_zoom (const gchar *file_name,
  * Loads a new #GdkPixbuf from @file_name and returns it.  This pixbuf is scaled
  * from the size indicated to the new size indicated by @width and @height.  If
  * either of these are -1, then the default size of the image being loaded is
- * used.  The caller must assume the reference to the reurned pixbuf. If an
+ * used.  The caller must assume the reference to the returned pixbuf. If an
  * error occurred, @error is set and %NULL is returned.
  * 
  * Return value: A newly allocated #GdkPixbuf, or %NULL
@@ -1679,7 +1830,7 @@ rsvg_pixbuf_from_file_at_size (const gchar *file_name,
 			       GError     **error)
 {
   FILE *f;
-  char chars[10];
+  char chars[SVG_BUFFER_SIZE];
   gint result;
   GdkPixbuf *retval;
   RsvgHandle *handle;
@@ -1692,14 +1843,16 @@ rsvg_pixbuf_from_file_at_size (const gchar *file_name,
       return NULL;
     }
   handle = rsvg_handle_new ();
-  data.zoom_set = FALSE;
+  data.type = RSVG_SIZE_WH;
   data.width = width;
   data.height = height;
 
   rsvg_handle_set_size_callback (handle, rsvg_size_callback, &data, NULL);
-  while ((result = fread (chars, 1, 3, f)) > 0)
+  while ((result = fread (chars, 1, SVG_BUFFER_SIZE, f)) > 0)
     rsvg_handle_write (handle, chars, result, error);
 
+  rsvg_handle_close (handle, error);
+  
   retval = rsvg_handle_get_pixbuf (handle);
 
   fclose (f);
@@ -1707,3 +1860,55 @@ rsvg_pixbuf_from_file_at_size (const gchar *file_name,
   return retval;
 }
 
+/**
+ * rsvg_pixbuf_from_file_at_max_size:
+ * @file_name: A file name
+ * @max_width: The requested max width
+ * @max_height: The requested max heigh
+ * @error: return location for errors
+ * 
+ * Loads a new #GdkPixbuf from @file_name and returns it.  This pixbuf is uniformly
+ * scaled so that the it fits into a rectangle of size max_width * max_height. The
+ * caller must assume the reference to the returned pixbuf. If an error occurred,
+ * @error is set and %NULL is returned.
+ * 
+ * Return value: A newly allocated #GdkPixbuf, or %NULL
+ **/
+GdkPixbuf  *
+rsvg_pixbuf_from_file_at_max_size (const gchar     *file_name,
+				   gint             max_width,
+				   gint             max_height,
+				   GError         **error)
+{
+  FILE *f;
+  char chars[SVG_BUFFER_SIZE];
+  gint result;
+  GdkPixbuf *retval;
+  RsvgHandle *handle;
+  struct RsvgSizeCallbackData data;
+
+  f = fopen (file_name, "r");
+  if (!f)
+    {
+      /* FIXME: Set up error. */
+      return NULL;
+    }
+  handle = rsvg_handle_new ();
+  data.type = RSVG_SIZE_WH_MAX;
+  data.width = max_width;
+  data.height = max_height;
+
+  rsvg_handle_set_size_callback (handle, rsvg_size_callback, &data, NULL);
+  while ((result = fread (chars, 1, SVG_BUFFER_SIZE, f)) > 0)
+    rsvg_handle_write (handle, chars, result, error);
+
+  rsvg_handle_close (handle, error);
+  
+  retval = rsvg_handle_get_pixbuf (handle);
+
+  fclose (f);
+  rsvg_handle_free (handle);
+  return retval;
+}
+
+
